require(serialize)
require(gdbm)
envlist <- function(e) {
    names <- ls(e, all=TRUE)
    list <- vector("list",length(names))
    for (i in seq(along=names))
        list[[i]] <- get(names[i], env=e, inherits=FALSE)
    names(list) <- names
    list
}
listIntoEnv <- function(list, e) {
    names <- names(list)
    for (i in seq(along = names))
        assign(names[i], list[[i]], env = e)
}
envtable <- function() {
    idx <- 0
    envs <- NULL
    enames <- character(0)
    find <- function(v, keys, vals)
        for (i in seq(along=keys))
            if (identical(v, keys[[i]]))
                return(vals[i])
    getname <- function(e) find(e, envs, enames)
    getenv <- function(n) find(n, enames, envs)
    insert <- function(e) {
        idx <<- idx + 1
        name <- paste("env", idx, sep="::")
        envs <<- c(e, envs)
        enames <<- c(name, enames)
        name
    }
    list(insert = insert, getenv = getenv, getname = getname)
}
makeShelf <- function(list, file, ascii = TRUE)
{
    names <- names(list)
    if (length(names) != length(list))
        stop("must provide a named list of values")
    table <- envtable()
    ser <- function(val) serialize(val, NULL, ascii = ascii, refhook = envhook)
    envhook <- function(e) {
        if (is.environment(e)) {
            name <- table$getname(e)
            if (is.null(name)) {
                name <- table$insert(e)
                data <- list(bindings = envlist(e), enclos = parent.env(e))
                dbinsert(file, name, ser(data))
            }
            name
        }
    }
    db <- dbnew(file)
    for (i in seq(along=names)) {
        name <- paste("var", names[i], sep="::")
        dbinsert(file, name, ser(list[[i]]))
    }
}
listShelf <- function(file) {
    prefpat <- "^var::"
    sub(prefpat, "", grep(prefpat,dblist(file), value=TRUE))
}
openShelf <- function(file) {
    envenv <- new.env(hash = TRUE)
    varkey <- function(name) paste("var", name, sep="::")
    fetch <- function(name)
        unserialize(dbfetch(file, name), refhook = envhook)
    envhook <- function(n) {
       if (exists(n, env = envenv, inherits = FALSE))
           get(n, env = envenv, inherits = FALSE)
       else {
           e <- new.env(hash = TRUE)
           assign(n, e, env = envenv) # MUST do this immediately
           data <- fetch(n)
           parent.env(e) <- data$enclos
           listIntoEnv(data$bindings, e)
           e
        }
    }
    list(getvar = function(name) fetch(varkey(name)),
         exists = function(name) dbexists(file, varkey(name)),
         list = function() listShelf(file))
}
loadShelf <- function(file, envir = parent.frame()) {
    s <- openShelf(file)
    wrap<-function(name) {
        name <- name # need to force evaluation!
        delay(s$getvar(name), env=environment())
    }
    for (n in listShelf(file))
        assign(n, wrap(n), envir = envir)
}
attachShelf <- function(file, pos = 2, name = paste("shelf", file, sep=":")) {
    env <- attach(NULL, pos = pos, name = name)
    on.exit(detach(pos))
    loadShelf(file, env)
    on.exit()
}
